# Copyright © 2023-2024 ValidMind Inc. All rights reserved.
# See the LICENSE file in the root of this repository for details.
# SPDX-License-Identifier: AGPL-3.0 AND ValidMind Commercial

import warnings

import plotly.express as px
from datasets import Dataset

from validmind import tags, tasks
from validmind.errors import MissingDependencyError

from .utils import get_ragas_config, get_renamed_columns

try:
    from ragas import evaluate
    from ragas.metrics import noise_sensitivity_relevant
except ImportError as e:
    raise MissingDependencyError(
        "Missing required package `ragas` for NoiseSensitivity. "
        "Please run `pip install validmind[llm]` to use LLM tests",
        required_dependencies=["ragas"],
        extra="llm",
    ) from e


@tags("ragas", "llm", "rag_performance")
@tasks("text_qa", "text_generation", "text_summarization")
def NoiseSensitivity(
    dataset,
    answer_column="answer",
    contexts_column="contexts",
    ground_truth_column="ground_truth",
):
    """
    Assesses the sensitivity of a Large Language Model (LLM) to noise in retrieved context by measuring how often it
    generates incorrect responses.

    ### Purpose

    The Noise Sensitivity test aims to measure how sensitive an LLM is to irrelevant or noisy information within the
    contextual data used to generate its responses. A lower noise sensitivity score suggests better model robustness in
    generating accurate answers from given contexts.

    ### Test Mechanism

    This test evaluates the model's answers by comparing the claims made in the generated response against the ground
    truth and the retrieved context. The noise sensitivity score is calculated as:

    $$
    \\text{noise sensitivity} = {|\\text{Number of incorrect claims in answer}| \\over |\\text{Number of total claims in answer}|}
    $$

    The formula computes the fraction of incorrect claims to the total claims in the answer, using a dataset where
    'answer', 'context', and 'ground_truth' columns are specified.

    #### Configuring Columns

    This metric requires the following columns in your dataset:

    - `contexts` (List[str]): A list of text contexts which are retrieved to generate
    the answer.
    - `answer` (str): The response generated by the model
    - `ground_truth` (str): The "correct" answer to the question

    If the above data is not in the appropriate column, you can specify different column
    names for these fields using the parameters `contexts_column` and `answer_column`.

    For example, if your dataset has this data stored in different columns, you can
    pass the following parameters:
    ```python
    {
        "contexts_column": "context_info"
        "answer_column": "my_answer_col",
    }
    ```

    If the data is stored as a dictionary in another column, specify the column and key
    like this:
    ```python
    pred_col = dataset.prediction_column(model)
    params = {
        "contexts_column": f"{pred_col}.contexts",
        "answer_column": f"{pred_col}.answer",
    }
    ```

    For more complex situations, you can use a function to extract the data:
    ```python
    pred_col = dataset.prediction_column(model)
    params = {
        "contexts_column": lambda row: [row[pred_col]["context_message"]],
        "answer_column": lambda row: "\\n\\n".join(row[pred_col]["messages"]),
    }

    ### Signs of High Risk

    - High noise sensitivity scores across multiple samples.
    - Significant deviation between mean and median noise sensitivity scores.
    - High standard deviation indicating inconsistency in the model's performance.

    ### Strengths

    - Provides a quantitative measure of how well the LLM handles noisy or irrelevant context.
    - Easy integration and configuration using column parameters.
    - Utilizes both histogram and box plot visualizations to analyze score distribution.

    ### Limitations

    - Requires accurate ground truth that aligns with the generated answers.
    - Assumes the context provided is sufficiently granular to assess noise sensitivity.
    - Primarily applicable to tasks like text QA, text generation, and text summarization where contextual relevance is
    critical.
    """
    warnings.filterwarnings(
        "ignore",
        category=FutureWarning,
        message="promote has been superseded by promote_options='default'.",
    )

    required_columns = {
        "answer": answer_column,
        "contexts": contexts_column,
        "ground_truth": ground_truth_column,
    }

    df = get_renamed_columns(dataset._df, required_columns)

    result_df = evaluate(
        Dataset.from_pandas(df),
        metrics=[noise_sensitivity_relevant],
        **get_ragas_config(),
    ).to_pandas()

    fig_histogram = px.histogram(
        x=result_df["noise_sensitivity_relevant"].to_list(), nbins=10
    )
    fig_box = px.box(x=result_df["noise_sensitivity_relevant"].to_list())

    return (
        {
            # "Scores (will not be uploaded to UI)": result_df[
            #     ["contexts", "answer", "ground_truth", "noise_sensitivity_relevant"]
            # ],
            "Aggregate Scores": [
                {
                    "Mean Score": result_df["noise_sensitivity_relevant"].mean(),
                    "Median Score": result_df["noise_sensitivity_relevant"].median(),
                    "Max Score": result_df["noise_sensitivity_relevant"].max(),
                    "Min Score": result_df["noise_sensitivity_relevant"].min(),
                    "Standard Deviation": result_df["noise_sensitivity_relevant"].std(),
                    "Count": result_df.shape[0],
                }
            ],
        },
        fig_histogram,
        fig_box,
    )
